import json
import typing
from collections.abc import Iterable
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Final,
    Generic,
    Literal,
    Mapping,
    Optional,
    TypeAlias,
    TypeVar,
    Union,
    cast,
)


@dataclass(frozen=True)
class TypeDescriptor:
    type: "Type"
    records: tuple["Record", ...]

    def as_json(self) -> Any:
        return _TYPE_DESCRIPTOR_SERIALIZER.to_json(self)

    def as_json_code(self) -> str:
        return json.dumps(self.as_json(), indent=2)

    @staticmethod
    def from_json(json: Any) -> "TypeDescriptor":
        return _TYPE_DESCRIPTOR_SERIALIZER.from_json(json)

    @staticmethod
    def from_json_code(json_code: str) -> "TypeDescriptor":
        return _TYPE_DESCRIPTOR_SERIALIZER.from_json(json.loads(json_code))


@dataclass(frozen=True)
class PrimitiveType:
    kind: Literal["primitive"]
    value: Literal[
        "bool",
        "int32",
        "int64",
        "uint64",
        "float32",
        "float64",
        "timestamp",
        "string",
        "bytes",
    ]


@dataclass(frozen=True)
class OptionalType:
    kind: Literal["optional"]
    value: "Type"


@dataclass(frozen=True)
class ArrayType:
    kind: Literal["array"]

    @dataclass(frozen=True)
    class Array:
        item: "Type"
        key_chain: tuple[str, ...]

    value: Array


@dataclass(frozen=True)
class RecordType:
    kind: Literal["record"]
    value: str


Type: TypeAlias = Union[PrimitiveType, OptionalType, ArrayType, RecordType]


@dataclass(frozen=True)
class Field:
    name: str
    type: Optional[Type]
    number: int


@dataclass(frozen=True)
class Record:
    kind: Literal["struct", "enum"]
    id: str
    fields: tuple[Field, ...]
    removed_fields: tuple[int, ...]


# ==============================================================================
# INTERNAL: JSON serialization framework
# ==============================================================================


_T = TypeVar("_T")


@dataclass(frozen=True)
class _Serializer(Generic[_T]):
    to_json: Callable[[_T], Any]
    from_json: Callable[[Any], _T]


@dataclass(frozen=True)
class _FieldSerializer(Generic[_T]):
    name: str
    serializer: _Serializer[_T]
    default: Optional[_T] = None


def _primitive_serializer(check_type_fn: Callable[[Any], _T]) -> _Serializer[_T]:
    return _Serializer(check_type_fn, check_type_fn)


def _literal_union_serializer(s: set[_T]) -> _Serializer[_T]:
    def check_in(input: _T) -> _T:
        if input not in s:
            raise ValueError(f"{input} is not in {s}")
        return input

    return _Serializer(check_in, check_in)


def _dataclass_serializer(
    type: typing.Type[_T], fields: Iterable[_FieldSerializer]
) -> _Serializer[_T]:
    def to_json(obj: _T) -> dict[str, Any]:
        json = {}
        for field in fields:
            value = getattr(obj, field.name)
            if value != field.default:
                json[field.name] = field.serializer.to_json(value)
        return json

    def from_json(json: dict[str, Any]) -> _T:
        def field_from_json(field: _FieldSerializer) -> Any:
            value_json = json.get(field.name)
            if value_json is None:
                if field.default is not None:
                    return field.default
                else:
                    # Will raise an exception.
                    json[field.name]
            return field.serializer.from_json(value_json)

        return type(**{f.name: field_from_json(f) for f in fields})

    return _Serializer(to_json, from_json)


def _union_serializer(
    kind_to_serializer: Mapping[str, _Serializer],
) -> _Serializer:
    def to_json(input: Any) -> dict[str, Any]:
        kind = cast(Any, input).kind
        serializer = kind_to_serializer[kind]
        return serializer.to_json(input)

    def from_json(json: dict[str, Any]) -> Any:
        kind = json["kind"]
        serializer = kind_to_serializer[kind]
        return serializer.from_json(json)

    return _Serializer(to_json, from_json)


def _listuple_serializer(
    item_serializer: _Serializer[_T],
) -> _Serializer[tuple[_T, ...]]:
    def to_json(input: tuple[_T, ...]) -> list[Any]:
        return [item_serializer.to_json(e) for e in input]

    def from_json(json: list[Any]) -> tuple[_T, ...]:
        return tuple(item_serializer.from_json(e) for e in json)

    return _Serializer(to_json, from_json)


def _forwarding_serializer(
    get_serializer_fn: Callable[[], _Serializer[_T]],
) -> _Serializer[_T]:
    def to_json(input: _T) -> Any:
        return get_serializer_fn().to_json(input)

    def from_json(json: Any) -> _T:
        return get_serializer_fn().from_json(json)

    return _Serializer(to_json, from_json)


# ==============================================================================
# INTERNAL: JSON serialization of TypeDescriptor
# ==============================================================================


def _type_serializer() -> _Serializer[Type]:
    return _TYPE_SERIALIZER


_TYPE_SERIALIZER: Final = _union_serializer(
    {
        "primitive": _dataclass_serializer(
            PrimitiveType,
            [
                _FieldSerializer("kind", _literal_union_serializer({"primitive"})),
                _FieldSerializer(
                    "value",
                    _literal_union_serializer(
                        {
                            "bool",
                            "int32",
                            "int64",
                            "uint64",
                            "float32",
                            "float64",
                            "timestamp",
                            "string",
                            "bytes",
                        }
                    ),
                ),
            ],
        ),
        "optional": _dataclass_serializer(
            OptionalType,
            [
                _FieldSerializer(
                    "kind",
                    _literal_union_serializer({"optional"}),
                ),
                _FieldSerializer(
                    "value",
                    _forwarding_serializer(_type_serializer),
                ),
            ],
        ),
        "array": _dataclass_serializer(
            ArrayType,
            [
                _FieldSerializer(
                    "kind",
                    _literal_union_serializer({"array"}),
                ),
                _FieldSerializer(
                    "value",
                    _dataclass_serializer(
                        ArrayType.Array,
                        [
                            _FieldSerializer(
                                "item",
                                _forwarding_serializer(_type_serializer),
                            ),
                            _FieldSerializer(
                                "key_chain",
                                _listuple_serializer(_primitive_serializer(str)),
                                default=(),
                            ),
                        ],
                    ),
                ),
            ],
        ),
        "record": _dataclass_serializer(
            RecordType,
            [
                _FieldSerializer(
                    "kind",
                    _literal_union_serializer({"record"}),
                ),
                _FieldSerializer(
                    "value",
                    _primitive_serializer(str),
                ),
            ],
        ),
    }
)


_FIELD_SERIALIZER: Final = _dataclass_serializer(
    Field,
    [
        _FieldSerializer(
            "name",
            _primitive_serializer(str),
        ),
        _FieldSerializer(
            "type",
            _forwarding_serializer(_type_serializer),
        ),
        _FieldSerializer(
            "number",
            _primitive_serializer(int),
        ),
    ],
)


_RECORD_SERIALIZER: Final = _dataclass_serializer(
    Record,
    [
        _FieldSerializer(
            "kind",
            _literal_union_serializer({"struct", "enum"}),
        ),
        _FieldSerializer(
            "id",
            _primitive_serializer(str),
        ),
        _FieldSerializer(
            "fields",
            _listuple_serializer(_FIELD_SERIALIZER),
        ),
        _FieldSerializer(
            "removed_fields",
            _listuple_serializer(_primitive_serializer(int)),
            default=(),
        ),
    ],
)


_TYPE_DESCRIPTOR_SERIALIZER = _dataclass_serializer(
    TypeDescriptor,
    [
        _FieldSerializer(
            "type",
            _TYPE_SERIALIZER,
        ),
        _FieldSerializer(
            "records",
            _listuple_serializer(_RECORD_SERIALIZER),
        ),
    ],
)



from dataclasses import dataclass
from typing import Generic, TypeVar

from soia._impl.serializer import Serializer

Request = TypeVar("Request")
Response = TypeVar("Response")


@dataclass(frozen=True)
class Method(Generic[Request, Response]):
    """Identifies a procedure (the "P" in "RPC") on both client side and server side."""

    name: str
    number: int
    request_serializer: Serializer[Request]
    response_serializer: Serializer[Response]



from collections.abc import Callable
from typing import Protocol

from soia._impl.function_maker import ExprLike

from soia import _spec, reflection


class TypeAdapter(Protocol):
    def default_expr(self) -> ExprLike:
        """
        The default value for T.
        """
        ...

    def to_frozen_expr(self, arg_expr: ExprLike) -> ExprLike:
        """
        Transforms the argument passed to the constructor of a frozen class into a
        frozen value which will be assigned to the attribute of the frozen object.

        The type of the returned expression must be T even if the argument does not have
        the expected type. Ideally, the expression should raise an error in the latter
        case.
        """
        ...

    def is_not_default_expr(self, arg_expr: ExprLike, attr_expr: ExprLike) -> ExprLike:
        """
        Returns an expression which evaluates to true if the given value is *not* the
        default value for T.
        This expression is inserted in the constructor of a frozen class, after the
        attribute has been assigned from the result of freezing arg_expr.
        If possible, an implemtation should try to use arg_expr instead of attr_expr as
        it offers a marginal performance advantage ('x' vs 'self.x').
        """
        ...

    def to_json_expr(self, in_expr: ExprLike, readable: bool) -> ExprLike:
        """
        Returns an expression which can be passed to 'json.dumps()' in order to
        serialize the given T into JSON format.
        The JSON flavor (dense versus readable) is given by the 'readable' arg.
        """
        ...

    def from_json_expr(self, json_expr: ExprLike) -> ExprLike:
        """
        Transforms 'json_expr' into a T.
        The 'json_expr' arg is obtained by calling 'json.loads()'.
        """
        ...

    def finalize(
        self,
        resolve_type_fn: Callable[[_spec.Type], "TypeAdapter"],
    ) -> None: ...

    def get_type(self) -> reflection.Type: ...

    def register_records(
        self,
        registry: dict[str, reflection.Record],
    ) -> None: ...



import dataclasses
import unittest
from datetime import datetime, timedelta, timezone
from typing import Any

from soia import Timestamp


class TimestampTestCase(unittest.TestCase):
    def test_getters(self):
        ts = Timestamp(unix_millis=200)
        self.assertEqual(ts.unix_millis, 200)
        self.assertEqual(ts.unix_seconds, 0.2)

    def test_from_unix_millis(self):
        ts = Timestamp.from_unix_millis(200)
        self.assertEqual(ts.unix_millis, 200)

    def test_from_unix_seconds(self):
        ts = Timestamp.from_unix_seconds(200)
        self.assertEqual(ts.unix_millis, 200000)

    def test_from_datetime(self):
        ts = Timestamp.from_datetime(datetime.fromtimestamp(200, tz=timezone.utc))
        self.assertEqual(ts.unix_millis, 200000)

    def test_epoch(self):
        self.assertEqual(Timestamp.EPOCH.unix_millis, 0)

    def round_millis(self):
        unix_millis: Any = 200.8
        ts = Timestamp(unix_millis=unix_millis)
        self.assertEqual(ts.unix_millis, 201)

    def test_lower_bound(self):
        ts = Timestamp(unix_millis=-(10**20))
        self.assertEqual(ts.unix_millis, -8640000000000000)
        self.assertEqual(ts.unix_millis, Timestamp.MIN.unix_millis)

    def test_upper_bound(self):
        ts = Timestamp(unix_millis=10**20)
        self.assertEqual(ts.unix_millis, 8640000000000000)
        self.assertEqual(ts.unix_millis, Timestamp.MAX.unix_millis)

    def test_now(self):
        ts = Timestamp.now()
        now_as_dt = datetime.now(tz=timezone.utc)
        self.assertGreater(ts.unix_seconds, now_as_dt.timestamp() - 100)
        self.assertLess(ts.unix_seconds, now_as_dt.timestamp() + 100)

    def test_eq(self):
        self.assertEqual(
            Timestamp.from_unix_millis(2),
            Timestamp.from_unix_millis(2),
        )
        self.assertNotEqual(
            Timestamp.from_unix_millis(2),
            Timestamp.from_unix_millis(3),
        )

    def test_ordering(self):
        self.assertLess(
            Timestamp.from_unix_millis(2),
            Timestamp.from_unix_millis(3),
        )
        self.assertGreater(
            Timestamp.from_unix_millis(3),
            Timestamp.from_unix_millis(2),
        )

    def test_immutable(self):
        try:
            setattr(Timestamp.MIN, "unix_millis", 3)
            self.fail("Expected to fail with FrozenInstanceError")
        except dataclasses.FrozenInstanceError:
            pass

    def test_to_datetime(self):
        ts = Timestamp.from_unix_seconds(200)
        self.assertEqual(
            ts.to_datetime_or_raise(),
            datetime.fromtimestamp(200, tz=timezone.utc),
        )

    def test_to_datetime_out_of_bound(self):
        try:
            Timestamp.MIN.to_datetime_or_raise()
            self.fail("Expected to fail with OverflowError or ValueError")
        except Exception:
            pass

    def test_add_timedelta(self):
        ts = Timestamp.from_unix_seconds(200)
        actual: Timestamp = ts + timedelta(seconds=2, microseconds=3)
        self.assertEqual(actual, Timestamp.from_unix_seconds(202))

    def test_subtract_timedelta(self):
        ts = Timestamp.from_unix_seconds(200)
        actual: Timestamp = ts - timedelta(seconds=2, microseconds=3)
        self.assertEqual(actual, Timestamp.from_unix_seconds(198))

    def test_subtract_timestamp(self):
        a = Timestamp.from_unix_seconds(200)
        b = Timestamp.from_unix_seconds(104)
        actual: timedelta = a - b
        self.assertEqual(actual, timedelta(seconds=96))

    def test_repr(self):
        a = Timestamp.from_unix_seconds(123456789)
        self.assertEqual(
            repr(a),
            "Timestamp(\n  unix_millis=123456789000,\n  _formatted='1973-11-29T21:33:09Z',\n)",
        )
        self.assertEqual(
            str(a),
            "Timestamp(\n  unix_millis=123456789000,\n  _formatted='1973-11-29T21:33:09Z',\n)",
        )

    def test_repr_when_out_of_bounds(self):
        self.assertEqual(
            repr(Timestamp.MIN),
            "Timestamp(unix_millis=-8640000000000000)",
        )
